#ifndef _HSTLIB_H_
#define _HSTLIB_H_

#include "project.def"
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include "srvcore.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#ifndef MAX_HISTORY_RECORDS
# define MAX_HISTORY_RECORDS 100
#endif
#define HISTORY_RECORD_CAPACITY_ADD 100
#define POI_RANGE_FACTOR 0.10
#define POI_TOLERANCE_FACTOR 10.0
#define USE_POI_RANGE 1
#define USE_POI_TOLERANCE 2

typedef struct
{
  CONTRACT c;
  int  pollingRate; /* polling rate in msec */
  int  archiveRate; /* archive rate in msec */
  int  depthShort; /* for short term storage */
  int  depthLong; /* for long term storage */
  int  dataPtr; /* current valid data pointer */
  int  heartbeat; /* archive heartbeat in sec */
  float pTolerance; /* percent tolerance */
  float aTolerance; /* absolute tolerance */
  float range; /* value range from property settings */
  int  recordIndex; /* local index identifier */
  int  recordLength;
  int  recordInitialized;
  int  isScheduled; /* schedule flag */
  int  keepPointsOfInterest; /* poi flag */
  int  useStandardSet; /* use std. pre-allocated files */
  double lastArchiveTime;
  char archiveFileName[32]; /* current working file name */
  double *timestamp; /* timestamp array */
  SINT32 *sysstamp; /* sysstamp array */
  SINT32 *usrstamp; /* usrstamp array */
  PrpRedirBlk *prd; /* re-director to CHS if not NULL */
  BYTE *data;
} HstTblEntry;

typedef struct HistRecInputStruct
{
  CONTRACT c;
  int  pollingRate;               /* polling rate in msec */
  int  archiveRate;               /* archive rate in msec */
  int  depthShort;             /* for short term storage */
  int  depthLong;               /* for long term storage */
  int  heartbeat;            /* archive heartbeat in sec */
  int  recordIndex;            /* local index identifier */
  char exportTarget[EXPORT_NAME_SIZE];
  char redirection[PROPERTY_REDIR_SIZE];/* re-director to CHS if not NULL */
  char fmt[64];            /* format descriptor */
  char tolerance[64];      /* parseable tolerance string */
} HRIS;

/**
 * \struct HRS
 * \brief history record structure used in stock properties
 *
 * - property is the call property
 * - device is the call device name
 * - datasiz is the call size
 * - datafmt is the call format
 * - pollingRate is the polling rate in msec
 * - archiveRate is the archive rate in sec
 * - depthShort is the ring buffer depth for short term storage
 * - depthLong is the depth in months for long term disk storage
 * - heartbeat is the archive heartbeat in sec
 * - pTolerance is the percent tolerance filtering criterion
 * - aTolerance is the absolute tolerance filtering criterion
 */
typedef struct
{
  char property[PROPERTY_NAME_SIZE]; /* property name of the record */
  char device[DEVICE_NAME_SIZE]; /* device name of the record */
  int  datasiz;               /* call size of the record */
  int  datafmt;          /* calling format of the record */
  int  pollingRate;              /* polling rate in msec */
  int  archiveRate;              /* archive rate in msec */
  int  depthShort;             /* for short term storage */
  int  depthLong;               /* for long term storage */
  int  heartbeat;            /* archive heartbeat in sec */
  float pTolerance;                 /* percent tolerance */
  float aTolerance;                /* absolute tolerance */
} HRS; 
#define HRS_SIZE (64+9*4)

/**
 * \struct HistorySpecification
 * \brief Defines a TINE History Specification object
 *
 * - pollingRate is the polling rate in msec
 * - archiveRate is the archive rate in sec
 * - depthShort is the ring buffer depth for short term storage
 * - depthLong is the depth in months for long term disk storage
 * - heartbeat is the archive heartbeat in sec
 * - pTolerance is the percent tolerance filtering criterion
 * - aTolerance is the absolute tolerance filtering criterion
 * - rhsServerName is the Remote History Server Name
 * - rhsPropertyName is the associated archive property at the
 *   Remote History Server
 */
typedef struct
{
  int  pollingRate;              /* polling rate in msec */
  int  archiveRate;               /* archive rate in sec */
  int  depthShort;             /* for short term storage */
  int  depthLong;               /* for long term storage */
  int  heartbeat;            /* archive heartbeat in sec */
  float pTolerance;                 /* percent tolerance */
  float aTolerance;                /* absolute tolerance */
  char *rhsServerName;             /* Remote Server Name */
  char *rhsPropertyName;         /* Remote Property Name */
} HistorySpecification;

typedef struct
{
  double timeStamp;
  SINT32 sysStamp;
  SINT32 usrStamp;
} HstHdr;

#ifdef UNIX
# define SH_DENYNONE (S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH)
#endif
#ifdef VXWORKS
# undef SH_DENYNONE
# define SH_DENYNONE (0644)
#endif
#define HISTORY_FILENAME "history.csv"

#define HISTORY_BUFFER_SIZE 1024
#define STS_DEADBAND 1

extern char ArchiveLocation[HISTORY_HOME_SIZE];
extern HstTblEntry *hstTbl;
extern int nHistoryRecords;
extern TINE_EXPORT int minHistoryPollingRate;
extern TINE_EXPORT int maxHistoryRecords;
extern int gScavangerThreadBusy;
int scavengeOlderLTSFiles(UINT32 minDiskSpaceInBlocks,time_t currentFileTime);

/* useMinimalStorage */
/**
 * \brief Instructs the local history server to use a simple UTC timestamp 
 * (4-byte integer) when saving data to disk.
 *
 * By default, the local history subsystem will timestamp the data records stored to
 * disk with a higher resolution timestamp than the traditional UTC variable. The
 * data system stamps will likewise be stored with the data record.
 * This results in more disk space being used by an archive record, as the
 * record prefix is now 16 bytes per archived record instead of 4.  This can
 * be controlled by setting this variable to TRUE (small prefix) or FALSE (large prefix).
 *
 * \b Default: FALSE
 */
extern TINE_EXPORT int useMinimalStorage;
/* useMonthlyHistoryFiles */
/**
 * \brief Instructs the local history server to keep monthly archive files instead of daily
 * files.
 *
 * The local history subsystem maintains one file per record per day.  In some cases it 
 * might be advantageous to keep one file per record per month, for instance when many
 * small records are being maintained.  You can influence this behavior be setting this
 * flag to TRUE.
 */
extern TINE_EXPORT int useMonthlyHistoryFiles;

/* AppendHistoryInformation() */
/**
 * \brief Inserts a local history element into the local history server
 * 
 * A server can instruct the local history server to keep a history of the given
 * property by utilizing this call.  The local history server will periodically call
 * the property as specified according to the following input parameters.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param prp is the requested property for which a history is to be kept
 * \param dev is the device name to be associated with the property name supplied as the second
 *  parameter.
 * \param len is the length of the local history call.
 * \param fmt is the TINE format of the local history call.
 * \param idx is the local history index to be identified with this local history element 
 *  (Note: this must be unique with this server process and within the history data repository).
 * \param hspec is a pointer to a HistorySpecification structure, where the relevant filtering
 *  criteria are given.
 *  
 * \note In lieu of calling AppendHistoryInformation(), it is recommended to make use of the
 * startup configuration file 'history.csv', in which all local history critieria can be 
 * introduced.  This has the advantage that no information is hard-coded.  Indeed by introducing
 * a 'history.csv' file one can immediately instruct a server to maintain local history 
 * information (even if the server is developed and maintained by someone else!).
 *
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \b Example:
 *
 * \include eg_AppendHistoryInformation.c
 */
TINE_EXPORT int AppendHistoryInformation(char *eqm,char *prp,char *dev,int len,int fmt,int idx,HistorySpecification *hspec);
int LoadHistoryInformation(char *fname,char *expName,char *eqmName);
int getRemoteHistoryServer(char *eqm,char *prp);
UINT32 getAvailableDiskSpaceInBlocks(char *path);
void historyInit(char *expName,char *eqmName);
void historyCycle(int isScheduled);
int loadLTSData(HstTblEntry *hst,char *fn,int index,double dstarttime,double dendtime,
                 float lwrLimit, float uprLimit,
                 void *datbuf,int datfmt,char *dattag,int bufsize,int sample);
TINE_EXPORT int setSTSData(char *eqm,char *prp, char *dev,int index,BYTE *data,int fmt,int arraysize);
int getLTSData(char *eqm,char *prp, char *dev,int index,
               double dstarttime,double endtime,
               float lwrLimit, float uprLimit,
               BYTE *data,int fmt,char *tag,int arraysize,int sample,int recsreadFlag);
int getSTSData(char *eqm,char *prp, char *dev,int index,
               double dstarttime,double dendtime,
               float lwrLimit, float uprLimit,
               BYTE *data,int fmt,char *tag,int arraysize,int sample,int recsreadFlag);
void assignNumberOfPoints(BYTE *data,int fmt,int siz,int *npoints,int *nfound,int recsreadFlag);
int isInRange(DTYPE *d,float lwrLimit,float uprLimit);
time_t getStartTimeFromDepth(char *eqm,char *prp, int depth);

/* GetLastStoredData() */
/**
 * \brief Retreives the last stored data set from disk for the given input parameters
 * 
 * A server can retrieve the last stored data set from disk for any given local history
 * property.  This is frequently useful to insure that a server restart will reload data 
 * various sets, for instance when data are collected according to intermittent triggers, 
 * etc.
 *
 * \param eqm is the 6-character local equipment identifier name, which is internal to server.
 * \param prp is the requested property for which a history is to be kept
 * \param dev is the device name to be associated with the property name supplied as the second
 *  parameter.
 * \param idx is the local history index to be identified with this local history element 
 *  (Note: this must be unique with this server process and within the history data repository).
 * \param data is the buffer to receive the data.  It must be pre-allocated to at least the
 *  array length specified in the 'len' parameter.
 * \param fmt is the TINE format of the local history call.
 * \param len is the length of the local history call.
 *  
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \b Example:
 *
 * \include eg_getLastStoredData.c
 */
TINE_EXPORT int GetLastStoredData(char *eqm,char *prp, char *dev,int idx,BYTE *data,int fmt,int len);
#define getLastStoredData(a1,a2,a3,a4,a5,a6,a7) GetLastStoredData(a1,a2,a3,a4,a5,a6,a7)

TINE_EXPORT int GetLocalHistoryData(char *eqm,char *prp, char *dev,DTYPE *dout,DTYPE *din);
int HISTQueryFunction(CONTRACT *con,char *dataIn,char *data,double *dataTimeStamp);

/* SetMaximumHistoryTableRecords() */
/**
 * \brief Sets the maximum local history table size
 * 
 * The local history subsystem makes use of a table containing all relevant history contracts.
 * The maximum size of the this table is governed by this routine.
 *
 * \param value is the desired maximum history table size (default is 100).
 *  
 */
TINE_EXPORT void SetMaximumHistoryTableRecords(int value);

/* SetPointsOfInterestRangeFactor() */
/**
 * \brief Sets the definitive property range (maximum-minimum) factor which determines whether
 * a value is consider a point of interest.
 * 
 * If the change in a property's readback value is greater than the value's display range times
 * this range factor, then the value is considered a point of interest and therefore always
 * returned in archive calls over the relevant time range.
 *
 * \param value is the desired display range factor (default is 0.25).
 *  
 */
TINE_EXPORT void SetPointsOfInterestRangeFactor(float value);

/* SetPointsOfInterestToleranceFactor() */
/**
 * \brief Sets the definitive property tolerance factor which determines whether
 * a value is consider a point of interest.
 * 
 * If the change in a property's readback value is greater than the value's tolerance (if non-zero)
 * times this range factor, then the value is considered a point of interest and therefore always
 * returned in archive calls over the relevant time range.
 *
 * \param value is the desired display range factor (default is 10).
 *  
 */
TINE_EXPORT void SetPointsOfInterestToleranceFactor(float value);

/**
 * \internal
 */
int findAvailableHistoryIndex(void);

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* _HSTLIB_H_ */
